@@ -97,8 +97,8 @@ ldns_nsec_get_bitmap(ldns_rr *nsec) {
 /* this is NOT the hash, but the original name! */
 ldns_rdf *
 ldns_dnssec_nsec3_closest_encloser(ldns_rdf *qname,
-							ATTR_UNUSED(ldns_rr_type qtype),
-							ldns_rr_list *nsec3s)
+                                   ATTR_UNUSED(ldns_rr_type qtype),
+                                   ldns_rr_list *nsec3s)
 {
 	/* remember parameters, they must match */
 	uint8_t algorithm;
@@ -119,6 +119,7 @@ ldns_dnssec_nsec3_closest_encloser(ldns_rdf *qname,
 	size_t nsec_i;
 	ldns_rr *nsec;
 	ldns_rdf *result = NULL;
+	qtype = qtype;
 	
 	if (!qname || !nsec3s || ldns_rr_list_rr_count(nsec3s) < 1) {
 		return NULL;
@@ -547,8 +548,8 @@ break;
 
 ldns_rdf *
 ldns_dnssec_create_nsec_bitmap(ldns_rr_type rr_type_list[],
-						 size_t size,
-						 ldns_rr_type nsec_type)
+                               size_t size,
+                               ldns_rr_type nsec_type)
 {
 	size_t i;
 	uint8_t *bitmap;
@@ -585,15 +586,6 @@ ldns_dnssec_create_nsec_bitmap(ldns_rr_type rr_type_list[],
 				   (int) (7 - (i_type % 8)),
 				   true);
 	}
-	/* always add nsec (if this is not nsec3 and rrsig */
-	i_type = LDNS_RR_TYPE_RRSIG;
-	ldns_set_bit(bitmap + (int) i_type / 8, (int) (7 - (i_type % 8)), true);
-	i_type = nsec_type;
-	if (i_type != LDNS_RR_TYPE_NSEC3) {
-		ldns_set_bit(bitmap + (int) i_type / 8,
-				   (int) (7 - (i_type % 8)),
-				   true);
-	}
 
 	/* fold it into windows TODO: can this be done directly? */
 	memset(cur_data, 0, 32);
@@ -615,14 +607,13 @@ ldns_dnssec_create_nsec_bitmap(ldns_rr_type rr_type_list[],
 			cur_window++;
 			cur_window_max = 0;
 			memset(cur_data, 0, 32);
-		} else {
-			cur_data[i%32] = bitmap[i];
-			if (bitmap[i] > 0) {
-				cur_window_max = i%32;
-			}
+		}
+		cur_data[i%32] = bitmap[i];
+		if (bitmap[i] > 0) {
+			cur_window_max = i%32;
 		}
 	}
-	if (cur_window_max > 0) {
+	if (cur_window_max > 0 || cur_data[0] != 0) {
 		/* this window has stuff, add it */
 		data = LDNS_XREALLOC(data,
 						 uint8_t,
@@ -644,7 +635,9 @@ ldns_dnssec_create_nsec_bitmap(ldns_rr_type rr_type_list[],
 }
 
 ldns_rr *
-ldns_dnssec_create_nsec(ldns_dnssec_name *from, ldns_dnssec_name *to, ldns_rr_type nsec_type)
+ldns_dnssec_create_nsec(ldns_dnssec_name *from,
+                        ldns_dnssec_name *to,
+                        ldns_rr_type nsec_type)
 {
 	ldns_rr *nsec_rr;
 	ldns_rr_type types[1024];
@@ -667,10 +660,14 @@ ldns_dnssec_create_nsec(ldns_dnssec_name *from, ldns_dnssec_name *to, ldns_rr_ty
 		type_count++;
 		cur_rrsets = cur_rrsets->next;
 	}
+	types[type_count] = LDNS_RR_TYPE_RRSIG;
+	type_count++;
+	types[type_count] = LDNS_RR_TYPE_NSEC;
+	type_count++;
 
 	ldns_rr_push_rdf(nsec_rr, ldns_dnssec_create_nsec_bitmap(types,
-												  type_count,
-												  nsec_type));
+	                               type_count,
+	                               nsec_type));
 
 	return nsec_rr;
 }
@@ -699,18 +696,18 @@ ldns_dnssec_create_nsec3(ldns_dnssec_name *from,
 
 	nsec_rr = ldns_rr_new_frm_type(LDNS_RR_TYPE_NSEC3);
 	ldns_rr_set_owner(nsec_rr,
-				   ldns_nsec3_hash_name(ldns_dnssec_name_name(from),
-								    algorithm,
-								    iterations,
-								    salt_length,
-								    salt));
+	                  ldns_nsec3_hash_name(ldns_dnssec_name_name(from),
+	                  algorithm,
+	                  iterations,
+	                  salt_length,
+	                  salt));
 	status = ldns_dname_cat(ldns_rr_owner(nsec_rr), zone_name);
 	ldns_nsec3_add_param_rdfs(nsec_rr,
-						 algorithm,
-						 flags,
-						 iterations,
-						 salt_length,
-						 salt);
+	                          algorithm,
+	                          flags,
+	                          iterations,
+	                          salt_length,
+	                          salt);
 
 	cur_rrsets = from->rrsets;
 	while (cur_rrsets) {
@@ -718,20 +715,28 @@ ldns_dnssec_create_nsec3(ldns_dnssec_name *from,
 		type_count++;
 		cur_rrsets = cur_rrsets->next;
 	}
-
+	/* always add rrsig type if this is not an unsigned 
+	 * delegation
+	 */
+	if (type_count > 0 &&
+	    !(type_count == 1 && types[0] == LDNS_RR_TYPE_NS)) {
+		types[type_count] = LDNS_RR_TYPE_RRSIG;
+		type_count++;
+	}
+	
 	/* leave next rdata empty if they weren't precomputed yet */
 	if (to && to->hashed_name) {
 		ldns_rr_set_rdf(nsec_rr,
-					 ldns_rdf_clone(to->hashed_name),
-					 4);
+		                ldns_rdf_clone(to->hashed_name),
+		                4);
 	} else {
 		ldns_rr_set_rdf(nsec_rr, NULL, 4);
 	}
 
-	ldns_rr_set_rdf(nsec_rr,
-				 ldns_dnssec_create_nsec_bitmap(types,
-										  type_count,
-										  LDNS_RR_TYPE_NSEC3), 5);
+	ldns_rr_push_rdf(nsec_rr,
+	                 ldns_dnssec_create_nsec_bitmap(types,
+	                 type_count,
+	                 LDNS_RR_TYPE_NSEC3));
 
 	return nsec_rr;
 }
@@ -906,20 +911,21 @@ ldns_nsec3_hash_name(ldns_rdf *name,
 	hashed_owner_str_len = SHA_DIGEST_LENGTH;
 
 	hashed_owner_b32 = LDNS_XMALLOC(char,
-							  b32_ntop_calculate_size(
-							      hashed_owner_str_len) + 1);
+							  ldns_b32_ntop_calculate_size(
+							       hashed_owner_str_len) + 1);
 	hashed_owner_b32_len =
-		(size_t) b32_ntop_extended_hex((uint8_t *) hashed_owner_str,
-								 hashed_owner_str_len,
-								 hashed_owner_b32,
-								 b32_ntop_calculate_size(
-								     hashed_owner_str_len));
+		(size_t) ldns_b32_ntop_extended_hex((uint8_t *) hashed_owner_str,
+									 hashed_owner_str_len,
+									 hashed_owner_b32,
+									 ldns_b32_ntop_calculate_size(
+										 hashed_owner_str_len));
 	if (hashed_owner_b32_len < 1) {
 		fprintf(stderr, "Error in base32 extended hex encoding ");
 		fprintf(stderr, "of hashed owner name (name: ");
 		ldns_rdf_print(stderr, name);
 		fprintf(stderr, ", return code: %d)\n", hashed_owner_b32_len);
-		exit(4);
+		LDNS_FREE(hashed_owner_b32);
+		return NULL;
 	}
 	hashed_owner_str_len = hashed_owner_b32_len;
 	hashed_owner_b32[hashed_owner_b32_len] = '\0';
@@ -927,7 +933,8 @@ ldns_nsec3_hash_name(ldns_rdf *name,
 	status = ldns_str2rdf_dname(&hashed_owner, hashed_owner_b32);
 	if (status != LDNS_STATUS_OK) {
 		fprintf(stderr, "Error creating rdf from %s\n", hashed_owner_b32);
-		exit(1);
+		LDNS_FREE(hashed_owner_b32);
+		return NULL;
 	}
 
 	LDNS_FREE(hashed_owner_b32);
@@ -944,17 +951,19 @@ ldns_nsec3_add_param_rdfs(ldns_rr *rr,
 {
 	ldns_rdf *salt_rdf = NULL;
 	uint8_t *salt_data = NULL;
-	
+
 	ldns_rr_set_rdf(rr,
 				 ldns_rdf_new_frm_data(LDNS_RDF_TYPE_INT8,
 								   1,
 								   (void*)&algorithm),
 				 0);
+
 	ldns_rr_set_rdf(rr,
 				 ldns_rdf_new_frm_data(LDNS_RDF_TYPE_INT8,
 								   1,
 								   (void*)&flags),
 				 1);
+	
 	ldns_rr_set_rdf(rr,
 				 ldns_native2rdf_int16(LDNS_RDF_TYPE_INT16,
 								   iterations),
@@ -1001,6 +1010,7 @@ ldns_create_nsec3(ldns_rdf *cur_owner,
 	uint16_t cur_data_size = 0;
 
 	ldns_status status;
+
 	
 	hashed_owner = ldns_nsec3_hash_name(cur_owner,
 								 algorithm,
@@ -1108,11 +1118,10 @@ ldns_create_nsec3(ldns_rdf *cur_owner,
 		cur_data_size += cur_window_max + 3;
 	}
 
-	ldns_rr_set_rdf(nsec,
-				 ldns_rdf_new_frm_data(LDNS_RDF_TYPE_NSEC,
-								   cur_data_size,
-								   data),
-				 5);
+	ldns_rr_push_rdf(nsec,
+	                 ldns_rdf_new_frm_data(LDNS_RDF_TYPE_NSEC,
+	                                       cur_data_size,
+	                                       data));
 
 	LDNS_FREE(bitmap);
 	LDNS_FREE(data);
@@ -1137,7 +1146,7 @@ ldns_nsec3_flags(const ldns_rr *nsec3_rr)
 	if (nsec3_rr && ldns_rr_get_type(nsec3_rr) == LDNS_RR_TYPE_NSEC3 &&
 	    ldns_rdf_size(ldns_rr_rdf(nsec3_rr, 1)) > 0
 	    ) {
-		return ldns_rdf2native_int8(ldns_rr_rdf(nsec3_rr, 0));
+		return ldns_rdf2native_int8(ldns_rr_rdf(nsec3_rr, 1));
 	}
 	return 0;
 }
@@ -1536,61 +1545,6 @@ ldns_rr_list_sort_nsec3(ldns_rr_list *unsorted)
 	      qsort_rr_compare_nsec3);
 }
 
-ldns_status
-ldns_dnssec_zone_create_nsec3s(ldns_dnssec_zone *zone,
-						 ldns_rr_list *new_rrs,
-						 uint8_t algorithm,
-						 uint8_t flags,
-						 uint16_t iterations,
-						 uint8_t salt_length,
-						 uint8_t *salt)
-{
-	ldns_rbnode_t *first_name_node;
-	ldns_rbnode_t *current_name_node;
-	ldns_dnssec_name *current_name;
-	ldns_status result = LDNS_STATUS_OK;
-	ldns_rr *nsec_rr;
-	ldns_rr_list *nsec3_list;
-	
-	if (!zone || !new_rrs || !zone->names) {
-		return LDNS_STATUS_ERR;
-	}
-
-	nsec3_list = ldns_rr_list_new();
-
-	first_name_node = ldns_dnssec_name_node_next_nonglue(
-					  ldns_rbtree_first(zone->names));
-	
-	current_name_node = first_name_node;
-
-	while (current_name_node &&
-		  current_name_node != LDNS_RBTREE_NULL) {
-		current_name = (ldns_dnssec_name *) current_name_node->data;
-		nsec_rr = ldns_dnssec_create_nsec3(current_name,
-									NULL,
-								     zone->soa->name,
-									algorithm,
-									flags,
-									iterations,
-									salt_length,
-									salt);
-		ldns_dnssec_name_add_rr(current_name, nsec_rr);
-		ldns_rr_list_push_rr(new_rrs, nsec_rr);
-		ldns_rr_list_push_rr(nsec3_list, nsec_rr);
-		current_name_node = ldns_dnssec_name_node_next_nonglue(
-						    ldns_rbtree_next(current_name_node));
-	}
-
-	ldns_rr_list_sort_nsec3(nsec3_list);
-	ldns_dnssec_chain_nsec3_list(nsec3_list);
-	if (result != LDNS_STATUS_OK) {
-		return result;
-	}
-	
-	ldns_rr_list_free(nsec3_list);
-	return result;
-}
-
 int
 ldns_dnssec_default_add_to_signatures(ldns_rr *sig, void *n)
 {
