@@ -44,6 +44,8 @@ ldns_sign_public(ldns_rr_list *rrset, ldns_key_list *keys)
 	if (!rrset || ldns_rr_list_rr_count(rrset) < 1 || !keys) {
 		return NULL;
 	}
+	
+	new_owner = NULL;
 
 	key_count = 0;
 	signatures = ldns_rr_list_new();
@@ -93,7 +95,11 @@ ldns_sign_public(ldns_rr_list *rrset, ldns_key_list *keys)
 			printf("[XX]ERROR NO SIGN BUG, OUT OF MEM?\n");
 			ldns_rr_list_print(stdout, rrset_clone);
 			fflush(stdout);
-			exit(123);
+			ldns_rr_list_free(rrset_clone);
+			ldns_rr_list_free(signatures);
+			ldns_rdf_free(first_label);
+			ldns_rdf_free(new_owner);
+			return NULL;
 		}
 		b64rdf = NULL;
 
@@ -216,12 +222,14 @@ ldns_sign_public(ldns_rr_list *rrset, ldns_key_list *keys)
 				break;
 #ifdef USE_SHA2
 			case LDNS_SIGN_RSASHA256:
+			case LDNS_SIGN_RSASHA256_NSEC3:
 				b64rdf = ldns_sign_public_evp(
 						   sign_buf,
 						   ldns_key_evp_key(current_key),
 						   EVP_sha256());
 				break;
 			case LDNS_SIGN_RSASHA512:
+			case LDNS_SIGN_RSASHA512_NSEC3:
 				b64rdf = ldns_sign_public_evp(
 						   sign_buf,
 						   ldns_key_evp_key(current_key),
@@ -337,8 +345,9 @@ ldns_sign_public_evp(ldns_buffer *to_sign,
 	/* initializes a signing context */
 	md_type = digest_type;
 	if(!md_type) {
-		printf("Unknown message digest");
-		exit(1);
+		/* unknown message difest */
+		ldns_buffer_free(b64sig);
+		return NULL;
 	}
 
 	EVP_MD_CTX_init(&ctx);
@@ -457,8 +466,7 @@ ldns_dnssec_zone_mark_glue(ldns_dnssec_zone *zone)
 	while (cur_node != LDNS_RBTREE_NULL) {
 		cur_name = (ldns_dnssec_name *) cur_node->data;
 		cur_node = ldns_rbtree_next(cur_node);
-
-		if (cur_name->rrsets && !cur_name->rrsets->next &&
+		if (cur_name->rrsets && 
 		    (cur_name->rrsets->type == LDNS_RR_TYPE_A ||
 			cur_name->rrsets->type == LDNS_RR_TYPE_AAAA
 			)) {
@@ -470,6 +478,11 @@ ldns_dnssec_zone_mark_glue(ldns_dnssec_zone *zone)
 				if (ldns_dnssec_zone_find_rrset(zone,
 										  cur_owner,
 										  LDNS_RR_TYPE_NS)) {
+					/*
+					fprintf(stderr, "[XX] Marking as glue: ");
+					ldns_rdf_print(stderr, cur_name->name);
+					fprintf(stderr, "\n");
+					*/
 					cur_name->is_glue = true;
 				}
 				cur_parent = ldns_dname_left_chop(cur_owner);
@@ -510,13 +523,30 @@ ldns_dnssec_name_node_next_nonglue(ldns_rbnode_t *node)
 
 ldns_status
 ldns_dnssec_zone_create_nsecs(ldns_dnssec_zone *zone,
-						ldns_rr_list *new_rrs)
+                              ldns_rr_list *new_rrs)
 {
 
 	ldns_rbnode_t *first_node, *cur_node, *next_node;
 	ldns_dnssec_name *cur_name, *next_name;
 	ldns_rr *nsec_rr;
-
+	uint32_t nsec_ttl;
+	ldns_dnssec_rrsets *soa;
+	
+	/* the TTL of NSEC rrs should be set to the minimum TTL of
+	 * the zone SOA (RFC4035 Section 2.3)
+	 */
+	soa = ldns_dnssec_name_find_rrset(zone->soa, LDNS_RR_TYPE_SOA);
+	
+	/* did the caller actually set it? if not,
+	 * fall back to default ttl
+	 */
+	if (soa && soa->rrs && soa->rrs->rr) {
+		nsec_ttl = ldns_rdf2native_int32(ldns_rr_rdf(
+		                                     soa->rrs->rr, 6));
+	} else {
+		nsec_ttl = LDNS_DEFAULT_TTL;
+	}
+	
 	first_node = ldns_dnssec_name_node_next_nonglue(
 			       ldns_rbtree_first(zone->names));
 	cur_node = first_node;
@@ -531,8 +561,9 @@ ldns_dnssec_zone_create_nsecs(ldns_dnssec_zone *zone,
 		cur_name = (ldns_dnssec_name *)cur_node->data;
 		next_name = (ldns_dnssec_name *)next_node->data;
 		nsec_rr = ldns_dnssec_create_nsec(cur_name,
-								    next_name,
-								    LDNS_RR_TYPE_NSEC);
+		                                  next_name,
+		                                  LDNS_RR_TYPE_NSEC);
+		ldns_rr_set_ttl(nsec_rr, nsec_ttl);
 		ldns_dnssec_name_add_rr(cur_name, nsec_rr);
 		ldns_rr_list_push_rr(new_rrs, nsec_rr);
 		cur_node = next_node;
@@ -546,8 +577,9 @@ ldns_dnssec_zone_create_nsecs(ldns_dnssec_zone *zone,
 		cur_name = (ldns_dnssec_name *)cur_node->data;
 		next_name = (ldns_dnssec_name *)first_node->data;
 		nsec_rr = ldns_dnssec_create_nsec(cur_name,
-								    next_name,
-								    LDNS_RR_TYPE_NSEC);
+		                                  next_name,
+		                                  LDNS_RR_TYPE_NSEC);
+		ldns_rr_set_ttl(nsec_rr, nsec_ttl);
 		ldns_dnssec_name_add_rr(cur_name, nsec_rr);
 		ldns_rr_list_push_rr(new_rrs, nsec_rr);
 	} else {
@@ -557,6 +589,84 @@ ldns_dnssec_zone_create_nsecs(ldns_dnssec_zone *zone,
 	return LDNS_STATUS_OK;
 }
 
+ldns_status
+ldns_dnssec_zone_create_nsec3s(ldns_dnssec_zone *zone,
+						 ldns_rr_list *new_rrs,
+						 uint8_t algorithm,
+						 uint8_t flags,
+						 uint16_t iterations,
+						 uint8_t salt_length,
+						 uint8_t *salt)
+{
+	ldns_rbnode_t *first_name_node;
+	ldns_rbnode_t *current_name_node;
+	ldns_dnssec_name *current_name;
+	ldns_status result = LDNS_STATUS_OK;
+	ldns_rr *nsec_rr;
+	ldns_rr_list *nsec3_list;
+	uint32_t nsec_ttl;
+	ldns_dnssec_rrsets *soa;
+	
+	if (!zone || !new_rrs || !zone->names) {
+		return LDNS_STATUS_ERR;
+	}
+	
+	/* the TTL of NSEC rrs should be set to the minimum TTL of
+	 * the zone SOA (RFC4035 Section 2.3)
+	 */
+	soa = ldns_dnssec_name_find_rrset(zone->soa, LDNS_RR_TYPE_SOA);
+	
+	/* did the caller actually set it? if not,
+	 * fall back to default ttl
+	 */
+	if (soa && soa->rrs && soa->rrs->rr) {
+		nsec_ttl = ldns_rdf2native_int32(ldns_rr_rdf(
+		                                     soa->rrs->rr, 6));
+	} else {
+		nsec_ttl = LDNS_DEFAULT_TTL;
+	}
+
+	nsec3_list = ldns_rr_list_new();
+
+	first_name_node = ldns_dnssec_name_node_next_nonglue(
+					  ldns_rbtree_first(zone->names));
+	
+	current_name_node = first_name_node;
+
+	while (current_name_node &&
+	       current_name_node != LDNS_RBTREE_NULL) {
+		current_name = (ldns_dnssec_name *) current_name_node->data;
+		nsec_rr = ldns_dnssec_create_nsec3(current_name,
+		                                   NULL,
+		                                   zone->soa->name,
+		                                   algorithm,
+		                                   flags,
+		                                   iterations,
+		                                   salt_length,
+		                                   salt);
+		/* by default, our nsec based generator adds rrsigs
+		 * remove the bitmap for empty nonterminals */
+		if (!current_name->rrsets) {
+			ldns_rdf_deep_free(ldns_rr_pop_rdf(nsec_rr));
+		}
+		ldns_rr_set_ttl(nsec_rr, nsec_ttl);
+		ldns_dnssec_name_add_rr(current_name, nsec_rr);
+		ldns_rr_list_push_rr(new_rrs, nsec_rr);
+		ldns_rr_list_push_rr(nsec3_list, nsec_rr);
+		current_name_node = ldns_dnssec_name_node_next_nonglue(
+		                   ldns_rbtree_next(current_name_node));
+	}
+
+	ldns_rr_list_sort_nsec3(nsec3_list);
+	ldns_dnssec_chain_nsec3_list(nsec3_list);
+	if (result != LDNS_STATUS_OK) {
+		return result;
+	}
+	
+	ldns_rr_list_free(nsec3_list);
+	return result;
+}
+
 ldns_dnssec_rrs *
 ldns_dnssec_remove_signatures(ldns_dnssec_rrs *signatures,
 						ldns_key_list *key_list,
@@ -647,10 +757,10 @@ ldns_dnssec_remove_signatures(ldns_dnssec_rrs *signatures,
 
 ldns_status
 ldns_dnssec_zone_create_rrsigs(ldns_dnssec_zone *zone,
-						 ldns_rr_list *new_rrs,
-						 ldns_key_list *key_list,
-						 int (*func)(ldns_rr *, void*),
-						 void *arg)
+                               ldns_rr_list *new_rrs,
+                               ldns_key_list *key_list,
+                               int (*func)(ldns_rr *, void*),
+                               void *arg)
 {
 	ldns_status result = LDNS_STATUS_OK;
 
@@ -702,24 +812,30 @@ ldns_dnssec_zone_create_rrsigs(ldns_dnssec_zone *zone,
 					cur_rr = cur_rr->next;
 				}
 				
-				siglist = ldns_sign_public(rr_list, key_list);
-				for (i = 0; i < ldns_rr_list_rr_count(siglist); i++) {
-					if (cur_rrset->signatures) {
-						ldns_dnssec_rrs_add_rr(cur_rrset->signatures,
-										   ldns_rr_list_rr(siglist,
-													    i));
-					} else {
-						cur_rrset->signatures = ldns_dnssec_rrs_new();
-						cur_rrset->signatures->rr =
-							ldns_rr_list_rr(siglist, i);
-						ldns_rr_list_push_rr(new_rrs,
-										 ldns_rr_list_rr(siglist,
-													  i));
+				/* only sign non-delegation RRsets */
+				/* (glue should have been marked earlier) */
+				if (ldns_rr_list_type(rr_list) != LDNS_RR_TYPE_NS ||
+					ldns_dname_compare(ldns_rr_list_owner(rr_list),
+					zone->soa->name) == 0) {
+
+					siglist = ldns_sign_public(rr_list, key_list);
+					for (i = 0; i < ldns_rr_list_rr_count(siglist); i++) {
+						if (cur_rrset->signatures) {
+							ldns_dnssec_rrs_add_rr(cur_rrset->signatures,
+											   ldns_rr_list_rr(siglist,
+														    i));
+						} else {
+							cur_rrset->signatures = ldns_dnssec_rrs_new();
+							cur_rrset->signatures->rr =
+								ldns_rr_list_rr(siglist, i);
+							ldns_rr_list_push_rr(new_rrs,
+											 ldns_rr_list_rr(siglist,
+														  i));
+						}
 					}
+					ldns_rr_list_free(siglist);
 				}
 				
-				
-				ldns_rr_list_free(siglist);
 				ldns_rr_list_free(rr_list);
 				
 				cur_rrset = cur_rrset->next;
@@ -835,6 +951,9 @@ ldns_dnssec_zone_sign_nsec3(ldns_dnssec_zone *zone,
 									 iterations,
 									 salt_length,
 									 salt);
+				/* always set bit 7 of the flags to zero, according to
+				 * rfc5155 section 11 */
+				ldns_set_bit(ldns_rdf_data(ldns_rr_rdf(nsec3params, 1)), 7, 0);
 				ldns_dnssec_zone_add_rr(zone, nsec3params);
 				ldns_rr_list_push_rr(new_rrs, nsec3params);
 			}
