@@ -2,7 +2,7 @@
  * ldns-dpa inspects the (udp) DNS packets found in a pcap file
  * and provides statistics about them
  * 
- * (C) NLnet Labs 2006
+ * (C) NLnet Labs 2006 - 2008
  *
  * See the file LICENSE for the license
  */
@@ -10,6 +10,7 @@
 
 #include <ldns/ldns.h>
 
+#ifdef HAVE_PCAP_H
 #ifdef HAVE_LIBPCAP
 
 #ifdef HAVE_NETINET_IP6_H
@@ -87,6 +88,7 @@ enum enum_match_ids {
 	MATCH_DST_ADDRESS,
 	MATCH_TIMESTAMP,
 	MATCH_QUERY,
+	MATCH_QTYPE,
 	MATCH_ANSWER,
 	MATCH_AUTHORITY,
 	MATCH_ADDITIONAL,
@@ -103,6 +105,7 @@ enum enum_counter_types {
 	TYPE_TIMESTAMP,
 	TYPE_ADDRESS,
 	TYPE_RR,
+	TYPE_RR_TYPE,
 	TYPE_LAST
 };
 typedef enum enum_counter_types counter_type;
@@ -182,6 +185,7 @@ const type_operators const_type_operators[] = {
 	{ TYPE_TIMESTAMP, 6, { OP_EQUAL, OP_NOTEQUAL, OP_GREATER, OP_LESSER, OP_GREATEREQUAL, OP_LESSEREQUAL, 0, 0, 0, 0 } },
 	{ TYPE_ADDRESS, 3, { OP_EQUAL, OP_NOTEQUAL, OP_CONTAINS, 0, 0, 0, 0, 0, 0, 0} },
 	{ TYPE_RR, 3, { OP_EQUAL, OP_NOTEQUAL, OP_CONTAINS, 0, 0, 0, 0, 0, 0, 0} },
+	{ TYPE_RR_TYPE, 6, { OP_EQUAL, OP_NOTEQUAL, OP_GREATER, OP_LESSER, OP_GREATEREQUAL, OP_LESSEREQUAL, 0, 0, 0, 0 } },
 	{ 0, 0, { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 } }
 };
 
@@ -210,6 +214,7 @@ typedef struct struct_match_table match_table;
  */
 const match_table matches[] = {
 	{ MATCH_QUERY, "query", "String representation of the query RR", TYPE_RR },
+	{ MATCH_QTYPE, "qtype", "RR Type of the question RR, if present", TYPE_RR_TYPE },
 	{ MATCH_SRC_ADDRESS, "srcaddress", "address the packet was sent from", TYPE_ADDRESS },
 	{ MATCH_TIMESTAMP, "timestamp", "time the packet was sent", TYPE_TIMESTAMP },
 	{ MATCH_DST_ADDRESS, "dstaddress", "address the packet was sent to", TYPE_ADDRESS },
@@ -1016,6 +1021,41 @@ match_str(type_operator operator,
 	}	
 }
 
+bool
+match_rr_type(type_operator operator,
+              char *value,
+	      char *mvalue)
+{
+	ldns_rr_type a,b;
+	
+	a = ldns_get_rr_type_by_name(value);
+	b = ldns_get_rr_type_by_name(mvalue);
+	
+	switch (operator) {
+		case OP_EQUAL:
+			return a == b;
+			break;
+		case OP_NOTEQUAL:
+			return a != b;
+			break;
+		case OP_GREATER:
+			return a > b;
+			break;
+		case OP_LESSER:
+			return a < b;
+			break;
+		case OP_GREATEREQUAL:
+			return a >= b;
+			break;
+		case OP_LESSEREQUAL:
+			return a <= b;
+			break;
+		default:
+			fprintf(stderr, "Unknown operator: %u\n", operator);
+			exit(2);
+	}
+}
+
 bool
 match_rcode(type_operator operator,
              char *value,
@@ -1127,6 +1167,9 @@ value_matches(match_id id,
 		case MATCH_DST_ADDRESS:
 			result = match_str(operator, value, mvalue);
 			break;
+		case MATCH_QTYPE:
+			result = match_rr_type(operator, value, mvalue);
+			break;
 		default:
 			fprintf(stderr, "Error: value_matches() for operator %s not implemented yet.\n", get_op_str((type_operator) id));
 			exit(3);
@@ -1279,6 +1322,14 @@ get_string_value(match_id id, ldns_pkt *pkt, ldns_rdf *src_addr, ldns_rdf *dst_a
 				val[0] = '\0';
 			}
 			break;
+		case MATCH_QTYPE:
+			if (ldns_pkt_qdcount(pkt) > 0) {
+				free(val);
+				val = ldns_rr_type2str(ldns_rr_get_type(ldns_rr_list_rr(ldns_pkt_question(pkt), 0)));
+			} else {
+				val[0] = '\0';
+			}
+			break;
 		case MATCH_ANSWER:
 			if (ldns_pkt_ancount(pkt) > 0) {
 				free(val);
@@ -2315,7 +2366,7 @@ printf("timeval: %u ; %u\n", cur_hdr.ts.tv_sec, cur_hdr.ts.tv_usec);
 				printf("\t; ??? - %u\n", ip_len);
 */
 			} else {
-				printf("Lost fragment %u\n", iptr->ip_id);
+				/*printf("Lost fragment %u\n", iptr->ip_id);*/
 				lost_packet_fragments++;
 				return 1;
 			}
@@ -2988,7 +3039,13 @@ int main(int argc, char *argv[]) {
 
 #else
 int main() {
-	fprintf(stderr, "ldns-dpa was not built because there is no pcap library on this system, or i couldn't find it. Please install pcap and rebuild.\n");
+	fprintf(stderr, "ldns-dpa was not built because there is no pcap library on this system, or there was no pcap header file at compilation time. Please install pcap and rebuild.\n");
+	return 1;
+}
+#endif
+#else
+int main() {
+	fprintf(stderr, "ldns-dpa was not built because there is no pcap library on this system, or there was no pcap header file at compilation time. Please install pcap and rebuild.\n");
 	return 1;
 }
 #endif
