@@ -65,6 +65,7 @@ ldns_lookup_table ldns_rr_classes[] = {
         { LDNS_RR_CLASS_IN, "IN" },
         { LDNS_RR_CLASS_CH, "CH" },
         { LDNS_RR_CLASS_HS, "HS" },
+        { LDNS_RR_CLASS_NONE, "NONE" },
         { LDNS_RR_CLASS_ANY, "ANY" },
         { 0, NULL }
 };
@@ -94,6 +95,130 @@ ldns_lookup_table ldns_opcodes[] = {
         { 0, NULL }
 };
 
+ldns_status
+ldns_pkt_opcode2buffer_str(ldns_buffer *output, ldns_pkt_opcode opcode)
+{
+	ldns_lookup_table *lt = ldns_lookup_by_id(ldns_opcodes, opcode);
+	if (lt && lt->name) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "OPCODE%u", opcode);
+	}
+	return ldns_buffer_status(output);
+}
+
+ldns_status
+ldns_pkt_rcode2buffer_str(ldns_buffer *output, ldns_pkt_rcode rcode)
+{
+	ldns_lookup_table *lt = ldns_lookup_by_id(ldns_rcodes, rcode);
+	if (lt && lt->name) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "RCODE%u", rcode);
+	}
+	return ldns_buffer_status(output);
+}
+
+ldns_status
+ldns_algorithm2buffer_str(ldns_buffer *output,
+                          ldns_algorithm algorithm)
+{
+	ldns_lookup_table *lt = ldns_lookup_by_id(ldns_algorithms,
+	                                          algorithm);
+	if (lt && lt->name) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "ALG%u", algorithm);
+	}
+	return ldns_buffer_status(output);
+}
+
+ldns_status
+ldns_cert_algorithm2buffer_str(ldns_buffer *output,
+                               ldns_cert_algorithm cert_algorithm)
+{
+	ldns_lookup_table *lt = ldns_lookup_by_id(ldns_cert_algorithms,
+	                                          cert_algorithm);
+	if (lt && lt->name) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "CERT_ALG%u",
+		                   cert_algorithm);
+	}
+	return ldns_buffer_status(output);
+}
+
+char *
+ldns_pkt_opcode2str(ldns_pkt_rcode opcode)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(12);
+	str = NULL;
+
+	if (ldns_pkt_opcode2buffer_str(buf, opcode) == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+char *
+ldns_pkt_rcode2str(ldns_pkt_rcode rcode)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(10);
+	str = NULL;
+
+	if (ldns_pkt_rcode2buffer_str(buf, rcode) == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+char *
+ldns_pkt_algorithm2str(ldns_algorithm algorithm)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(10);
+	str = NULL;
+
+	if (ldns_algorithm2buffer_str(buf, algorithm)
+	    == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+char *
+ldns_pkt_cert_algorithm2str(ldns_algorithm cert_algorithm)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(10);
+	str = NULL;
+
+	if (ldns_cert_algorithm2buffer_str(buf, cert_algorithm)
+	    == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+
 /* do NOT pass compressed data here :p */
 ldns_status
 ldns_rdf2buffer_str_dname(ldns_buffer *output, const ldns_rdf *dname)
@@ -220,7 +345,7 @@ ldns_rdf2buffer_str_str(ldns_buffer *output, const ldns_rdf *rdf)
 	ldns_buffer_printf(output, "\"");
 	for (i = 1; i <= length; ++i) {
 		char ch = (char) data[i];
-		if (isprint(ch)) {
+		if (isprint((int)ch)) {
 			if (ch == '"' || ch == '\\') {
 				ldns_buffer_printf(output, "\\");
 			}
@@ -236,21 +361,23 @@ ldns_rdf2buffer_str_str(ldns_buffer *output, const ldns_rdf *rdf)
 ldns_status
 ldns_rdf2buffer_str_b64(ldns_buffer *output, const ldns_rdf *rdf)
 {
-	size_t size = b64_ntop_calculate_size(ldns_rdf_size(rdf));
+	size_t size = ldns_b64_ntop_calculate_size(ldns_rdf_size(rdf));
 	char *b64 = LDNS_XMALLOC(char, size);
-	if (b64_ntop(ldns_rdf_data(rdf), ldns_rdf_size(rdf), b64, size)) {
+	if (ldns_b64_ntop(ldns_rdf_data(rdf), ldns_rdf_size(rdf), b64, size)) {
 		ldns_buffer_printf(output, "%s", b64);
 	}
 	LDNS_FREE(b64);
 	return ldns_buffer_status(output);
-}	
+}
 
 ldns_status
 ldns_rdf2buffer_str_b32_ext(ldns_buffer *output, const ldns_rdf *rdf)
 {
-	size_t size = b32_ntop_calculate_size(ldns_rdf_size(rdf) - 1);
+	size_t size = ldns_b32_ntop_calculate_size(ldns_rdf_size(rdf) - 1);
 	char *b32 = LDNS_XMALLOC(char, size + 1);
-	size = (size_t) b32_ntop_extended_hex(ldns_rdf_data(rdf) + 1, ldns_rdf_size(rdf) - 1, b32, size);
+	size = (size_t) ldns_b32_ntop_extended_hex(ldns_rdf_data(rdf) + 1,
+									   ldns_rdf_size(rdf) - 1,
+									   b32, size);
 	if (size > 0) {
 		ldns_buffer_printf(output, "%s", b32);
 	}
@@ -287,7 +414,7 @@ ldns_rdf2buffer_str_type(ldns_buffer *output, const ldns_rdf *rdf)
 ldns_status
 ldns_rdf2buffer_str_class(ldns_buffer *output, const ldns_rdf *rdf)
 {
-        uint8_t data = ldns_rdf_data(rdf)[0];
+	uint16_t data = ldns_read_uint16(ldns_rdf_data(rdf));
 	ldns_lookup_table *lt;
 
  	lt = ldns_lookup_by_id(ldns_rr_classes, (int) data);
@@ -352,6 +479,88 @@ loc_cm_print(ldns_buffer *output, uint8_t mantissa, uint8_t exponent)
 		ldns_buffer_printf(output, "0");
 }
 
+ldns_status
+ldns_rr_type2buffer_str(ldns_buffer *output, const ldns_rr_type type)
+{
+	const ldns_rr_descriptor *descriptor;
+
+	descriptor = ldns_rr_descript(type);
+
+	if (descriptor && descriptor->_name) {
+		ldns_buffer_printf(output, "%s", descriptor->_name);
+	} else {
+		/* exceptions for pseudotypes */
+		switch (type) {
+			case LDNS_RR_TYPE_IXFR:
+				ldns_buffer_printf(output, "IXFR");
+				break;
+			case LDNS_RR_TYPE_AXFR:
+				ldns_buffer_printf(output, "AXFR");
+				break;
+			case LDNS_RR_TYPE_MAILA:
+				ldns_buffer_printf(output, "MAILA");
+				break;
+			case LDNS_RR_TYPE_MAILB:
+				ldns_buffer_printf(output, "MAILB");
+				break;
+			case LDNS_RR_TYPE_ANY:
+				ldns_buffer_printf(output, "ANY");
+				break;
+			default:
+				ldns_buffer_printf(output, "TYPE%u", type);
+		}
+	}
+	return ldns_buffer_status(output);
+}
+
+char *
+ldns_rr_type2str(const ldns_rr_type type)
+{
+	char *str;
+	ldns_buffer *buf;
+
+	buf = ldns_buffer_new(10);
+	str = NULL;
+
+	if (ldns_rr_type2buffer_str(buf, type) == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+
+	ldns_buffer_free(buf);
+	return str;
+}
+
+
+ldns_status
+ldns_rr_class2buffer_str(ldns_buffer *output,
+                         const ldns_rr_class klass)
+{
+	ldns_lookup_table *lt;
+
+	lt = ldns_lookup_by_id(ldns_rr_classes, klass);
+	if (lt) {
+		ldns_buffer_printf(output, "%s", lt->name);
+	} else {
+		ldns_buffer_printf(output, "CLASS%d", klass);
+	}
+	return ldns_buffer_status(output);
+}
+
+char *
+ldns_rr_class2str(const ldns_rr_class klass)
+{
+	ldns_buffer *buf;
+	char *str;
+
+	str = NULL;
+	buf = ldns_buffer_new(10);
+	if (ldns_rr_class2buffer_str(buf, klass) == LDNS_STATUS_OK) {
+		str = ldns_buffer2str(buf);
+	}
+	ldns_buffer_free(buf);
+	return str;
+}
+
 ldns_status
 ldns_rdf2buffer_str_loc(ldns_buffer *output, const ldns_rdf *rdf)
 {
@@ -654,13 +863,15 @@ ldns_status
 ldns_rdf2buffer_str_int16_data(ldns_buffer *output, const ldns_rdf *rdf)
 {
 	/* Subtract the size (2) of the number that specifies the length */
-	size_t size = b64_ntop_calculate_size(ldns_rdf_size(rdf) - 2);
+	size_t size = ldns_b64_ntop_calculate_size(ldns_rdf_size(rdf) - 2);
 	char *b64 = LDNS_XMALLOC(char, size);
 
 	ldns_buffer_printf(output, "%u ", ldns_rdf_size(rdf) - 2);
 	
 	if (ldns_rdf_size(rdf) > 2 &&
-	    b64_ntop(ldns_rdf_data(rdf) + 2, ldns_rdf_size(rdf) - 2, b64, size)) {
+	    ldns_b64_ntop(ldns_rdf_data(rdf) + 2,
+				   ldns_rdf_size(rdf) - 2,
+				   b64, size)) {
 		ldns_buffer_printf(output, "%s", b64);
 	}
 	LDNS_FREE(b64);
@@ -846,11 +1057,8 @@ ldns_rdf2buffer_str(ldns_buffer *buffer, const ldns_rdf *rdf)
 ldns_status
 ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 {
-	uint16_t i;
+	uint16_t i, flags;
 	ldns_status status = LDNS_STATUS_OK;
-	ldns_lookup_table *lt;
-	const ldns_rr_descriptor *descriptor;
-
 	if (!rr) {
 		ldns_buffer_printf(output, "(null)\n");
 	} else {
@@ -868,36 +1076,12 @@ ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 		if (ldns_rr_rd_count(rr) > 0) {
 			ldns_buffer_printf(output, "\t%d", ldns_rr_ttl(rr));
 		}
-		
-		lt = ldns_lookup_by_id(ldns_rr_classes, ldns_rr_get_class(rr));
-		if (lt) {
-			ldns_buffer_printf(output, "\t%s\t", lt->name);
-		} else {
-			ldns_buffer_printf(output, "\tCLASS%d\t", 
-					ldns_rr_get_class(rr));
-		}
 
-		descriptor = ldns_rr_descript(ldns_rr_get_type(rr));
+		ldns_buffer_printf(output, "\t");
+		ldns_rr_class2buffer_str(output, ldns_rr_get_class(rr));
+		ldns_buffer_printf(output, "\t");
 
-		if (descriptor && descriptor->_name) {
-			ldns_buffer_printf(output, "%s", descriptor->_name);
-		} else {
-			/* exceptions for qtype */
-			if (ldns_rr_get_type(rr) == 251) {
-				ldns_buffer_printf(output, "IXFR ");
-			} else if (ldns_rr_get_type(rr) == 252) {
-				ldns_buffer_printf(output, "AXFR ");
-			} else if (ldns_rr_get_type(rr) == 253) {
-				ldns_buffer_printf(output, "MAILB ");
-			} else if (ldns_rr_get_type(rr) == 254) {
-				ldns_buffer_printf(output, "MAILA ");
-			} else if (ldns_rr_get_type(rr) == 255) {
-				ldns_buffer_printf(output, "ANY ");
-			} else {
-				ldns_buffer_printf(output, "TYPE%u", 
-						ldns_rr_get_type(rr));
-			}
-		}
+		ldns_rr_type2buffer_str(output, ldns_rr_get_type(rr));
 		
 		if (ldns_rr_rd_count(rr) > 0) {
 			ldns_buffer_printf(output, "\t");
@@ -906,7 +1090,7 @@ ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 			status = ldns_rdf2buffer_str(output, ldns_rr_rdf(rr, i));
 			if (i < ldns_rr_rd_count(rr) - 1) {
 				ldns_buffer_printf(output, " ");
-			} 
+			}
 		}
 		/* per RR special comments - handy for DNSSEC types */
 		/* check to prevent question sec. rr from
@@ -915,14 +1099,15 @@ ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 			switch (ldns_rr_get_type(rr)) {
 				case LDNS_RR_TYPE_DNSKEY:
 #ifdef HAVE_SSL
-					if (ldns_rdf2native_int16(ldns_rr_rdf(rr, 0)) == 256) {
+					flags = ldns_rdf2native_int16(ldns_rr_rdf(rr, 0));
+					if (flags == 256 || flags == 384) {
 						ldns_buffer_printf(output, 
 								" ;{id = %d (zsk), size = %db}", 
 								ldns_calc_keytag(rr),
 								ldns_rr_dnskey_key_size(rr)); 
 						break;
 					} 
-					if (ldns_rdf2native_int16(ldns_rr_rdf(rr, 0)) == 257) {
+					if (flags == 257 || flags == 385) {
 						ldns_buffer_printf(output, 
 								" ;{id = %d (ksk), size = %db}", 
 								ldns_calc_keytag(rr),
@@ -937,7 +1122,21 @@ ldns_rr2buffer_str(ldns_buffer *output, const ldns_rr *rr)
 				case LDNS_RR_TYPE_RRSIG:
 					ldns_buffer_printf(output, " ;{id = %d}", 
 							ldns_rdf2native_int16(ldns_rr_rdf(rr, 6)));
-							break;
+					break;
+				case LDNS_RR_TYPE_DS:
+					{
+						uint8_t *data = ldns_rdf_data(ldns_rr_rdf(rr, 3));
+						size_t len = ldns_rdf_size(ldns_rr_rdf(rr, 3));
+						char *babble = ldns_bubblebabble(data, len);
+						ldns_buffer_printf(output, " ; %s", babble);
+						LDNS_FREE(babble);
+					}
+					break;
+				case LDNS_RR_TYPE_NSEC3:
+					if (ldns_nsec3_optout(rr)) {
+						ldns_buffer_printf(output, " ; flags: optout");
+					}
+					break;
 				default:
 					break;
 
@@ -1125,6 +1324,24 @@ ldns_pkt2buffer_str(ldns_buffer *output, const ldns_pkt *pkt)
 	return status;
 }
 
+static ldns_status
+ldns_hmac_key2buffer_str(ldns_buffer *output, const ldns_key *k)
+{
+	ldns_status status;
+	size_t i;
+	ldns_rdf *b64_bignum;
+	
+	ldns_buffer_printf(output, "Key: ");
+
+ 	i = ldns_key_hmac_size(k);
+	b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, ldns_key_hmac_key(k));
+	status = ldns_rdf2buffer_str(output, b64_bignum);
+	ldns_rdf_deep_free(b64_bignum);
+	ldns_buffer_printf(output, "\n");
+	return status;
+}
+
+
 ldns_status
 ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 {
@@ -1155,12 +1372,14 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 			case LDNS_SIGN_RSASHA1:
 			case LDNS_SIGN_RSASHA1_NSEC3:
 			case LDNS_SIGN_RSASHA256:
+			case LDNS_SIGN_RSASHA256_NSEC3:
 			case LDNS_SIGN_RSASHA512:
+			case LDNS_SIGN_RSASHA512_NSEC3:
 			case LDNS_SIGN_RSAMD5:
 				/* copied by looking at dnssec-keygen output */
 				/* header */
 				rsa = ldns_key_rsa_key(k);
-				
+
 				ldns_buffer_printf(output,"Private-key-format: v1.2\n");
 				switch(ldns_key_algorithm(k)) {
 				case LDNS_SIGN_RSAMD5:
@@ -1169,22 +1388,36 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 								    LDNS_RSAMD5);
 					break;
 				case LDNS_SIGN_RSASHA1:
-				case LDNS_SIGN_RSASHA1_NSEC3:
 					ldns_buffer_printf(output,
 								    "Algorithm: %u (RSASHA1)\n",
 								    LDNS_RSASHA1);
 					break;
+				case LDNS_SIGN_RSASHA1_NSEC3:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA1_NSEC3)\n",
+								    LDNS_RSASHA1_NSEC3);
+					break;
 #ifdef USE_SHA2
 				case LDNS_SIGN_RSASHA256:
 					ldns_buffer_printf(output,
 								    "Algorithm: %u (RSASHA256)\n",
 								    LDNS_RSASHA256);
 					break;
+				case LDNS_SIGN_RSASHA256_NSEC3:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA256_NSEC3)\n",
+								    LDNS_RSASHA256_NSEC3);
+					break;
 				case LDNS_SIGN_RSASHA512:
 					ldns_buffer_printf(output,
 								    "Algorithm: %u (RSASHA512)\n",
 								    LDNS_RSASHA512);
 					break;
+				case LDNS_SIGN_RSASHA512_NSEC3:
+					ldns_buffer_printf(output,
+								    "Algorithm: %u (RSASHA512_NSEC3)\n",
+								    LDNS_RSASHA512_NSEC3);
+					break;
 #endif
 				default:
 					fprintf(stderr, "Warning: unknown signature ");
@@ -1324,7 +1557,11 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 				dsa = ldns_key_dsa_key(k);
 			
 				ldns_buffer_printf(output,"Private-key-format: v1.2\n");
-				ldns_buffer_printf(output,"Algorithm: 3 (DSA)\n");
+				if (ldns_key_algorithm(k) == LDNS_SIGN_DSA) {
+					ldns_buffer_printf(output,"Algorithm: 3 (DSA)\n");
+				} else if (ldns_key_algorithm(k) == LDNS_SIGN_DSA_NSEC3) {
+					ldns_buffer_printf(output,"Algorithm: 6 (DSA_NSEC3)\n");
+				}
 
 				/* print to buf, convert to bin, convert to b64,
 				 * print to buf */
@@ -1409,17 +1646,21 @@ ldns_key2buffer_str(ldns_buffer *output, const ldns_key *k)
 				}
 				break;
 			case LDNS_SIGN_HMACMD5:
-				/* is the filefmt specified for TSIG.. don't know */
-		                ldns_buffer_printf(output, "Private-key-format: v1.2\n");
-		                ldns_buffer_printf(output, "Algorithm: 157 (HMAC_MD5)\n");
-		                ldns_buffer_printf(output, "Key: ");
-		                i = ldns_key_hmac_size(k);
-				b64_bignum =  ldns_rdf_new_frm_data(LDNS_RDF_TYPE_B64, i, ldns_key_hmac_key(k));
-				if (ldns_rdf2buffer_str(output, b64_bignum) != LDNS_STATUS_OK) {
-					goto error;
-				}
-				ldns_rdf_deep_free(b64_bignum);
-				ldns_buffer_printf(output, "\n"); 
+				/* there's not much of a format defined for TSIG */
+				/* It's just a binary blob, Same for all algorithms */
+                ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+                ldns_buffer_printf(output, "Algorithm: 157 (HMAC_MD5)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
+				break;
+			case LDNS_SIGN_HMACSHA1:
+		        ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+		        ldns_buffer_printf(output, "Algorithm: 158 (HMAC_SHA1)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
+				break;
+			case LDNS_SIGN_HMACSHA256:
+		        ldns_buffer_printf(output, "Private-key-format: v1.2\n");
+		        ldns_buffer_printf(output, "Algorithm: 158 (HMAC_SHA256)\n");
+				status = ldns_hmac_key2buffer_str(output, k);
 				break;
 		}
 #endif /* HAVE_SSL */
@@ -1446,7 +1687,7 @@ error:
  * Zero terminate the buffer and fix it to the size of the string.
  */
 char *
-buffer2str(ldns_buffer *buffer)
+ldns_buffer2str(ldns_buffer *buffer)
 {
 	char *tmp_str;
 	char *str;
@@ -1478,7 +1719,7 @@ ldns_rdf2str(const ldns_rdf *rdf)
 
 	if (ldns_rdf2buffer_str(tmp_buffer, rdf) == LDNS_STATUS_OK) {
 		/* export and return string, destroy rest */
-		result = buffer2str(tmp_buffer);
+		result = ldns_buffer2str(tmp_buffer);
 	}
 	
 	ldns_buffer_free(tmp_buffer);
@@ -1493,9 +1734,8 @@ ldns_rr2str(const ldns_rr *rr)
 
 	if (ldns_rr2buffer_str(tmp_buffer, rr) == LDNS_STATUS_OK) {
 		/* export and return string, destroy rest */
-		result = buffer2str(tmp_buffer);
+		result = ldns_buffer2str(tmp_buffer);
 	}
-	
 	ldns_buffer_free(tmp_buffer);
 	return result;
 }
@@ -1508,7 +1748,7 @@ ldns_pkt2str(const ldns_pkt *pkt)
 
 	if (ldns_pkt2buffer_str(tmp_buffer, pkt) == LDNS_STATUS_OK) {
 		/* export and return string, destroy rest */
-		result = buffer2str(tmp_buffer);
+		result = ldns_buffer2str(tmp_buffer);
 	}
 
 	ldns_buffer_free(tmp_buffer);
@@ -1522,7 +1762,7 @@ ldns_key2str(const ldns_key *k)
 	ldns_buffer *tmp_buffer = ldns_buffer_new(LDNS_MIN_BUFLEN);
 	if (ldns_key2buffer_str(tmp_buffer, k) == LDNS_STATUS_OK) {
 		/* export and return string, destroy rest */
-		result = buffer2str(tmp_buffer);
+		result = ldns_buffer2str(tmp_buffer);
 	}
 	ldns_buffer_free(tmp_buffer);
 	return result;
@@ -1542,7 +1782,7 @@ ldns_rr_list2str(const ldns_rr_list *list)
 	}
 
 	/* export and return string, destroy rest */
-	result = buffer2str(tmp_buffer);
+	result = ldns_buffer2str(tmp_buffer);
 	ldns_buffer_free(tmp_buffer);
 	return result;
 }
