@@ -72,9 +72,12 @@ struct delegpt* delegpt_copy(struct delegpt* dp, struct regional* region)
 		if(!delegpt_add_ns(copy, region, ns->name))
 			return NULL;
 		copy->nslist->resolved = ns->resolved;
+		copy->nslist->got4 = ns->got4;
+		copy->nslist->got6 = ns->got6;
 	}
 	for(a = dp->target_list; a; a = a->next_target) {
-		if(!delegpt_add_addr(copy, region, &a->addr, a->addrlen))
+		if(!delegpt_add_addr(copy, region, &a->addr, a->addrlen, 
+			a->bogus, 0))
 			return NULL;
 	}
 	return copy;
@@ -107,6 +110,8 @@ delegpt_add_ns(struct delegpt* dp, struct regional* region, uint8_t* name)
 	dp->nslist = ns;
 	ns->name = regional_alloc_init(region, name, ns->namelen);
 	ns->resolved = 0;
+	ns->got4 = 0;
+	ns->got6 = 0;
 	return 1;
 }
 
@@ -124,25 +129,52 @@ delegpt_find_ns(struct delegpt* dp, uint8_t* name, size_t namelen)
 	return NULL;
 }
 
+struct delegpt_addr*
+delegpt_find_addr(struct delegpt* dp, struct sockaddr_storage* addr, 
+	socklen_t addrlen)
+{
+	struct delegpt_addr* p = dp->target_list;
+	while(p) {
+		if(sockaddr_cmp_addr(addr, addrlen, &p->addr, p->addrlen)==0) {
+			return p;
+		}
+		p = p->next_target;
+	}
+	return NULL;
+}
+
 int 
 delegpt_add_target(struct delegpt* dp, struct regional* region, 
 	uint8_t* name, size_t namelen, struct sockaddr_storage* addr, 
-	socklen_t addrlen)
+	socklen_t addrlen, int bogus, int nodup)
 {
 	struct delegpt_ns* ns = delegpt_find_ns(dp, name, namelen);
 	if(!ns) {
 		/* ignore it */
 		return 1;
 	}
-	ns->resolved = 1;
-	return delegpt_add_addr(dp, region, addr, addrlen);
+	if(addr_is_ip6(addr, addrlen))
+		ns->got6 = 1;
+	else	ns->got4 = 1;
+	if(ns->got4 && ns->got6)
+		ns->resolved = 1;
+	return delegpt_add_addr(dp, region, addr, addrlen, bogus, nodup);
 }
 
 int 
 delegpt_add_addr(struct delegpt* dp, struct regional* region, 
-	struct sockaddr_storage* addr, socklen_t addrlen)
+	struct sockaddr_storage* addr, socklen_t addrlen, int bogus, int nodup)
 {
-	struct delegpt_addr* a = (struct delegpt_addr*)regional_alloc(region,
+	struct delegpt_addr* a;
+	if(nodup) {
+		if((a = delegpt_find_addr(dp, addr, addrlen))) {
+			if(bogus)
+				a->bogus = bogus;
+			return 1;
+		}
+	}
+
+	a = (struct delegpt_addr*)regional_alloc(region,
 		sizeof(struct delegpt_addr));
 	if(!a)
 		return 0;
@@ -154,11 +186,11 @@ delegpt_add_addr(struct delegpt* dp, struct regional* region,
 	memcpy(&a->addr, addr, addrlen);
 	a->addrlen = addrlen;
 	a->attempts = 0;
+	a->bogus = bogus;
 	return 1;
 }
 
-/** count NS and number missing */
-static void
+void
 delegpt_count_ns(struct delegpt* dp, size_t* numns, size_t* missing)
 {
 	struct delegpt_ns* ns;
@@ -171,8 +203,7 @@ delegpt_count_ns(struct delegpt* dp, size_t* numns, size_t* missing)
 	}
 }
 
-/** count addresses, and number in result and available lists */
-static void
+void
 delegpt_count_addr(struct delegpt* dp, size_t* numaddr, size_t* numres, 
 	size_t* numavail)
 {
@@ -213,10 +244,15 @@ void delegpt_log(enum verbosity_value v, struct delegpt* dp)
 	if(verbosity >= VERB_ALGO) {
 		for(ns = dp->nslist; ns; ns = ns->next) {
 			dname_str(ns->name, buf);
-			log_info("  %s%s", buf, (ns->resolved?"*":""));
+			log_info("  %s %s%s%s%s", buf, (ns->resolved?"*":""),
+			(ns->got4?" A":""), (ns->got6?" AAAA":""),
+			(dp->bogus?" BOGUS":"") );
 		}
 		for(a = dp->target_list; a; a = a->next_target) {
-			log_addr(VERB_ALGO, "  ", &a->addr, a->addrlen);
+			if(a->bogus) 
+				log_addr(VERB_ALGO, "  BOGUS ", 
+					&a->addr, a->addrlen);
+			else log_addr(VERB_ALGO, "  ", &a->addr, a->addrlen);
 		}
 	}
 }
@@ -295,10 +331,10 @@ delegpt_from_message(struct dns_msg* msg, struct regional* region)
 			continue;
 
 		if(ntohs(s->rk.type) == LDNS_RR_TYPE_A) {
-			if(!delegpt_add_rrset_A(dp, region, s))
+			if(!delegpt_add_rrset_A(dp, region, s, 0))
 				return NULL;
 		} else if(ntohs(s->rk.type) == LDNS_RR_TYPE_AAAA) {
-			if(!delegpt_add_rrset_AAAA(dp, region, s))
+			if(!delegpt_add_rrset_AAAA(dp, region, s, 0))
 				return NULL;
 		}
 	}
@@ -312,6 +348,8 @@ delegpt_rrset_add_ns(struct delegpt* dp, struct regional* region,
 	struct packed_rrset_data* nsdata = (struct packed_rrset_data*)
 		ns_rrset->entry.data;
 	size_t i;
+	if(nsdata->security == sec_status_bogus)
+		dp->bogus = 1;
 	for(i=0; i<nsdata->count; i++) {
 		if(nsdata->rr_len[i] < 2+1) continue; /* len + root label */
 		if(dname_valid(nsdata->rr_data[i]+2, nsdata->rr_len[i]-2) !=
@@ -326,7 +364,7 @@ delegpt_rrset_add_ns(struct delegpt* dp, struct regional* region,
 
 int 
 delegpt_add_rrset_A(struct delegpt* dp, struct regional* region,
-	struct ub_packed_rrset_key* ak)
+	struct ub_packed_rrset_key* ak, int nodup)
 {
         struct packed_rrset_data* d=(struct packed_rrset_data*)ak->entry.data;
         size_t i;
@@ -341,7 +379,7 @@ delegpt_add_rrset_A(struct delegpt* dp, struct regional* region,
                 memmove(&sa.sin_addr, d->rr_data[i]+2, INET_SIZE);
                 if(!delegpt_add_target(dp, region, ak->rk.dname,
                         ak->rk.dname_len, (struct sockaddr_storage*)&sa,
-                        len))
+                        len, (d->security==sec_status_bogus), nodup))
                         return 0;
         }
         return 1;
@@ -349,7 +387,7 @@ delegpt_add_rrset_A(struct delegpt* dp, struct regional* region,
 
 int 
 delegpt_add_rrset_AAAA(struct delegpt* dp, struct regional* region,
-	struct ub_packed_rrset_key* ak)
+	struct ub_packed_rrset_key* ak, int nodup)
 {
         struct packed_rrset_data* d=(struct packed_rrset_data*)ak->entry.data;
         size_t i;
@@ -364,7 +402,7 @@ delegpt_add_rrset_AAAA(struct delegpt* dp, struct regional* region,
                 memmove(&sa.sin6_addr, d->rr_data[i]+2, INET6_SIZE);
                 if(!delegpt_add_target(dp, region, ak->rk.dname,
                         ak->rk.dname_len, (struct sockaddr_storage*)&sa,
-                        len))
+                        len, (d->security==sec_status_bogus), nodup))
                         return 0;
         }
         return 1;
@@ -379,9 +417,9 @@ delegpt_add_rrset(struct delegpt* dp, struct regional* region,
 	if(ntohs(rrset->rk.type) == LDNS_RR_TYPE_NS)
 		return delegpt_rrset_add_ns(dp, region, rrset);
 	else if(ntohs(rrset->rk.type) == LDNS_RR_TYPE_A)
-		return delegpt_add_rrset_A(dp, region, rrset);
+		return delegpt_add_rrset_A(dp, region, rrset, 0);
 	else if(ntohs(rrset->rk.type) == LDNS_RR_TYPE_AAAA)
-		return delegpt_add_rrset_AAAA(dp, region, rrset);
+		return delegpt_add_rrset_AAAA(dp, region, rrset, 0);
 	log_warn("Unknown rrset type added to delegpt");
 	return 1;
 }
