@@ -57,6 +57,7 @@
 #include "testcode/ldns-testpkts.h"
 #include "util/log.h"
 #include <signal.h>
+struct worker;
 
 /** Global variable: the scenario. Saved here for when event_init is done. */
 static struct replay_scenario* saved_scenario = NULL;
@@ -595,7 +596,7 @@ listen_delete(struct listen_dnsport* listen)
 }
 
 struct comm_base* 
-comm_base_create()
+comm_base_create(int ATTR_UNUSED(sigs))
 {
 	/* we return the runtime structure instead. */
 	struct replay_runtime* runtime = (struct replay_runtime*)
@@ -1160,4 +1161,8 @@ struct event_base* comm_base_internal(struct comm_base* ATTR_UNUSED(b))
 	return NULL;
 }
 
+void daemon_remote_exec(struct worker* ATTR_UNUSED(worker))
+{
+}
+
 /*********** End of Dummy routines ***********/
