@@ -136,7 +136,7 @@ comm_base_now(struct comm_base* b)
 #endif /* USE_MINI_EVENT */
 
 struct comm_base* 
-comm_base_create()
+comm_base_create(int sigs)
 {
 	struct comm_base* b = (struct comm_base*)calloc(1,
 		sizeof(struct comm_base));
@@ -148,10 +148,20 @@ comm_base_create()
 		return NULL;
 	}
 #ifdef USE_MINI_EVENT
+	(void)sigs;
 	/* use mini event time-sharing feature */
 	b->eb->base = event_init(&b->eb->secs, &b->eb->now);
 #else
+#  ifdef HAVE_EV_LOOP
+	/* libev */
+	if(sigs)
+		b->eb->base=(struct event_base *)ev_default_loop(EVFLAG_AUTO);
+	else
+		b->eb->base=(struct event_base *)ev_loop_new(EVFLAG_AUTO);
+#  else
+	(void)sigs;
 	b->eb->base = event_init();
+#  endif
 #endif
 	if(!b->eb->base) {
 		free(b->eb);
@@ -225,7 +235,10 @@ comm_point_send_udp_msg(struct comm_point *c, ldns_buffer* packet,
 {
 	ssize_t sent;
 	log_assert(c->fd != -1);
-	log_assert(ldns_buffer_remaining(packet) > 0);
+#ifdef UNBOUND_DEBUG
+	if(ldns_buffer_remaining(packet) == 0)
+		log_err("internal error: send empty UDP packet");
+#endif
 	log_assert(addr && addrlen > 0);
 	sent = sendto(c->fd, ldns_buffer_begin(packet), 
 		ldns_buffer_remaining(packet), 0,
@@ -554,15 +567,15 @@ int comm_point_perform_accept(struct comm_point* c,
 	if(new_fd == -1) {
 #ifndef USE_WINSOCK
 		/* EINTR is signal interrupt. others are closed connection. */
-		if(	errno != EINTR 
+		if(	errno == EINTR || errno == EAGAIN
 #ifdef EWOULDBLOCK
-			&& errno != EWOULDBLOCK 
+			|| errno == EWOULDBLOCK 
 #endif
 #ifdef ECONNABORTED
-			&& errno != ECONNABORTED 
+			|| errno == ECONNABORTED 
 #endif
 #ifdef EPROTO
-			&& errno != EPROTO
+			|| errno == EPROTO
 #endif /* EPROTO */
 			)
 			return -1;
@@ -1553,6 +1566,9 @@ comm_timer_set(struct comm_timer* timer, struct timeval* tv)
 		comm_timer_disable(timer);
 	event_set(&timer->ev_timer->ev, -1, EV_PERSIST|EV_TIMEOUT,
 		comm_timer_callback, timer);
+	if(event_base_set(timer->ev_timer->base->eb->base, 
+		&timer->ev_timer->ev) != 0)
+		log_err("comm_timer_set: set_base failed.");
 	if(evtimer_add(&timer->ev_timer->ev, tv) != 0)
 		log_err("comm_timer_set: evtimer_add failed.");
 	timer->ev_timer->enabled = 1;
